from pysat.solvers import Glucose3
from pysat.formula import CNF, IDPool  # for keeping track of variables
from pysat.card import CardEnc, EncType
from itertools import combinations
import time
import random
from statistics import mean
from datetime import datetime
import argparse


# Read input strings
argParser = argparse.ArgumentParser()
argParser.add_argument("-num", "--num_strings", type=int, help="number of strings")
argParser.add_argument("-len", "--len_strings", type=int, help="length of strings")
argParser.add_argument("-o", "--outfile", help="name of the output file")
args = argParser.parse_args()


# Input: Array of strings
# Output: Returns True if all strings have equal length
def equal_lengths(strings):
    length = len(strings[0])
    for sting in strings:
        if not len(sting) == length:
            return False
    return True


# Input: Array of strings and lambda function for the binary strings
# Output: SAT clauses to constraint the binary input strings in the model; A->1, C->2, G->3, T->4
def string_constraints(strings, binary_strings):
    curr_clauses = []
    for n in range(len(strings)):
        for l in range(len(strings[0])):
            if strings[n][l] == 'A':
                curr_clauses.append([binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'C':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'G':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'T':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([binary_strings(n, l, 3)])
            else:
                raise TypeError(f'String "{strings[n]}" contains a letter different from "A", "C", "G" and "T"')
    return curr_clauses


# Input: List of literals and a bound
# Output: Creates clauses to make sure that at most "bound" literals are True
def atmost_all_bounds(lits, bound):
    clauses = []
    for comb in list(combinations(lits, bound + 1)):
        clause = []
        for el in comb:
            clause.append(-el)
        clauses.append(clause)
    return clauses


# Input: List of all input strings; List with maximal distance for each string
# Output: new closest String or False if the instance is unsatisfiable
# With 6 strings of length 20 and d=10 it already takes a while...
def slim(strings, d):
    vpool = IDPool()
    clauses = []

    equal_lengths(strings)
    N = len(strings)
    L = len(strings[0])

    if not len(d) == N:
        raise TypeError(f'd contains {len(d)} distances, but {N} strings are given"')

    binary_strings = lambda n, l, k: vpool.id('binary_strings{0}-{1}-{2}'.format(n, l, k))

    # Define clauses for string array
    clauses.extend(string_constraints(strings, binary_strings))

    # closestString stores the currently closest string (binary encoded)
    closestString = lambda l, k: vpool.id('closestString{0}-{1}'.format(l, k))

    for l in range(L):
        cnf = CardEnc.equals(lits=[closestString(l, k) for k in range(4)], encoding=EncType.pairwise, bound=1)
        clauses.extend(cnf.clauses)

    # hammingDistances stores whether there is a distance between string n and the currently CS at position l
    # It might happen that too many entries are true, but since it is a satisfaction problem it does not matters
    hammingDistances = lambda n, l: vpool.id('hammingDistances{0}-{1}'.format(n, l))

    # Put distances into matrix
    for n in range(N):
        for l in range(L):
            for k in range(4):
                # When string and the current closest string are not the same at a position
                # then the hamming distance has to increase
                a = binary_strings(n, l, k)
                b = closestString(l, k)
                c = hammingDistances(n, l)
                # (c = 0) iff (a == b)
                # (a v b v c) /\ (-a v -b v c) /\ (a v -b v -c) /\ (-a v b v -c)
                clauses.extend([[-a, b, c], [a, -b, c]])

    # Constraint: All hamming distances smaller than d
    for n in range(N):
        cnf = atmost_all_bounds(lits=[hammingDistances(n, l) for l in range(L)], bound=d[n])
        clauses.extend(cnf)

    g = Glucose3()
    for clause in clauses:
        g.add_clause(clause)

    if g.solve():
        # print("This instance is solvable")
        model = g.get_model()

        cs = ""
        for l in range(L):
            if closestString(l, 0) in model:
                cs += "A"
            if closestString(l, 1) in model:
                cs += "C"
            if closestString(l, 2) in model:
                cs += "G"
            if closestString(l, 3) in model:
                cs += "T"
        # print(cs, "is the currently closestString")
        return cs
    else:
        return False


# Input: Two strings
# Output: Hamming distance between those two stings
def hamming_distance_2(string1, string2):
    if not len(string1) == len(string2):
        raise TypeError(f'Strings "{string1}" and "{string2}" do not have the same length')
    length = len(string1)
    dist = 0
    for i in range(length):
        if not string1[i] == string2[i]:
            dist += 1
    return dist


# Input: Closest String, List of input Strings, Boolean use_print
# Output: Returns distances of all Strings to the closest string.
# If use_print is True, the evaluation is printed
def evaluate(cs, strings, use_print):
    distances = []
    for n in range(len(strings)):
        distance = hamming_distance_2(cs, strings[n])
        if use_print:
            print("String", n, "has a hamming distance of", distance, "to closest String")
        distances.append(distance)
    return distances


# Input: length
# Output: Outputs a random String of given length, containing of the letters A, C, G, and T
def get_random_string(length):
    # choose from all lowercase letter
    letters = ["A", "C", "G", "T"]
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


if __name__ == '__main__':
    if args.num_strings is None:
        raise TypeError("You need to give the number of strings")
    else:
        num_strings = args.num_strings

    if args.len_strings is None:
        raise TypeError("You need to give the length of strings")
    else:
        len_strings = args.len_strings

    if args.outfile is None:
        raise TypeError("You need to give the name of the output file")
    else:
        outfile = args.outfile

    start_distance = 8
    num_results = 10  # Number of needed output results for sat and unsat instances each

    counter_sat = 0
    counter_unsat = 0

    time_sol = []
    time_no_sol = []

    distances = [start_distance] * num_strings

    with open(outfile, 'w') as f:
        while counter_sat < num_results or counter_unsat < num_results:
            # Create random strings
            strings = []
            for k in range(num_strings):
                strings.append(get_random_string(len_strings))

            try:
                # Do SLIM and measure time
                start = time.time()
                closestString = slim(strings, distances)
                end = time.time()

                if not closestString:
                    f.write(str(distances) + " - UNSAT" + "\n")
                    time_no_sol.append(end - start)
                    counter_unsat += 1
                    distances[random.randint(0, num_strings - 1)] += 1
                else:
                    f.write(str(distances) + " - SATISFIABLE" + "\n")
                    time_sol.append(end - start)
                    counter_sat += 1
                    distances[random.randint(0, num_strings - 1)] -= 1

            except Exception:
                f.write(str(distances) + " - EXCEPTION" + "\n")
                distances[random.randint(0, num_strings - 1)] += 1

        f.write("time_sol:" + str(time_sol) + "\n")
        f.write("Min time sol: " + str(min(time_sol)) + "\n")
        f.write("Mean time sol: " + str(mean(time_sol)) + "\n")
        f.write("Max time sol: " + str(max(time_sol)) + "\n")
        f.write("time_no_sol:" + str(time_no_sol) + "\n")
        f.write("Min time no_sol: " + str(min(time_no_sol)) + "\n")
        f.write("Mean time no_sol: " + str(mean(time_no_sol)) + "\n")
        f.write("Max time no_sol: " + str(max(time_no_sol)) + "\n")

    now = datetime.now()
    with open("output_times.txt", "a") as myfile:
        myfile.write(outfile + " - end time: " + now.strftime("%d/%m/%Y %H:%M:%S") + "\n")
        myfile.write(outfile + " - time_sol:" + str(time_sol) + "\n")
        myfile.write(outfile + " - time_no_sol:" + str(time_no_sol) + "\n")
