from pysat.solvers import Glucose3
from pysat.formula import CNF, IDPool  # for keeping track of variables
from pysat.card import CardEnc, EncType
from itertools import combinations
import numpy as np
import random
import argparse
from ctypes import c_char_p
import time
from multiprocessing import Process, Manager
from pathlib import Path
import logging

# Read input strings
argParser = argparse.ArgumentParser()
argParser.add_argument("-d", "--d", type=int, help="maximal distance d")
argParser.add_argument("-l", "--l", type=int, help="size of local instances")
argParser.add_argument("-its", "--iterations", type=int, help="number of iterations")
argParser.add_argument("-lt", "--localTimeout", type=float, help="local timeout")
argParser.add_argument("-str", "--strings", help="name of the input file with all strings")
argParser.add_argument("-cs", "--closestString", help="name of the input file with the currently closest string")
argParser.add_argument("-out", "--outfile", help="name of the output file")
args = argParser.parse_args()


# Input: Array of strings
# Output: Returns True if all strings have equal length
def equal_lengths(strings):
    length = len(strings[0])
    for sting in strings:
        if not len(sting) == length:
            return False
    return True


# Input: Two strings
# Output: Hamming distance between those two stings
def hamming_distance_2(string1, string2):
    if not len(string1) == len(string2):
        raise TypeError(f'Strings "{string1}" and "{string2}" do not have the same length')
    length = len(string1)
    dist = 0
    for i in range(length):
        if not string1[i] == string2[i]:
            dist += 1
    return dist


# Input: Array of strings and string s
# Output: Returns the maximum hamming distance between s and the strings in the array
def hamming_distance(strings, s):
    num_strings = len(strings)
    max_dist = 0
    for i in range(num_strings):
        try:
            curr_dist = hamming_distance_2(strings[i], s)
        except ValueError:
            raise
        max_dist = max(max_dist, curr_dist)
    return max_dist


# Input: Array of strings
# Output: Hamming diameter of the Strings (= maximal hamming distance between any two strings)
def hamming_diameter(strings):
    num_strings = len(strings)
    diameter = 0
    for i in range(num_strings):
        for j in range(i):
            try:
                curr_dist = hamming_distance_2(strings[j], strings[i])
            except ValueError:
                raise
            diameter = max(diameter, curr_dist)
    return diameter


# Input: Array of strings and lambda function for the binary strings
# Output: SAT clauses to constraint the binary input strings in the model; A->1, C->2, G->3, T->4
def string_constraints(strings, binary_strings):
    curr_clauses = []
    for n in range(len(strings)):
        for l in range(len(strings[0])):
            if strings[n][l] == 'A':
                curr_clauses.append([binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'C':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'G':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([binary_strings(n, l, 2)])
                curr_clauses.append([-binary_strings(n, l, 3)])
            elif strings[n][l] == 'T':
                curr_clauses.append([-binary_strings(n, l, 0)])
                curr_clauses.append([-binary_strings(n, l, 1)])
                curr_clauses.append([-binary_strings(n, l, 2)])
                curr_clauses.append([binary_strings(n, l, 3)])
            else:
                raise TypeError(f'String "{strings[n]}" contains a letter different from "A", "C", "G" and "T"')
    return curr_clauses

# Input: List of literals and a bound
# Output: Creates clauses to make sure that at most "bound" literals are True
def atmost_all_bounds(lits, bound):
    clauses = []
    for comb in list(combinations(lits, bound + 1)):
        clause = []
        for el in comb:
            clause.append(-el)
        clauses.append(clause)
    return clauses


# Input: Closest String, List of input Strings, Boolean use_print
# Output: Returns distances of all Strings to the closest string.
# If use_print is True, the evaluation is printed
def evaluate(cs, strings, use_print):
    distances = []
    for n in range(len(strings)):
        distance = hamming_distance_2(cs, strings[n])
        if use_print:
            print("String", n, "has a hamming distance of", distance, "to closest String")
        distances.append(distance)
    return distances


# Input: List of all input strings; List with maximal distance for each string
# Output: new closest String or False if the instance is unsatisfiable
# With 6 strings of length 20 and d=10 it already takes a while...
def slim(strings, d, slimClosestString):
    vpool = IDPool()
    clauses = []

    equal_lengths(strings)
    N = len(strings)
    L = len(strings[0])

    if not len(d) == N:
        raise TypeError(f'd contains {len(d)} distances, but {N} strings are given"')

    binary_strings = lambda n, l, k: vpool.id('binary_strings{0}-{1}-{2}'.format(n, l, k))

    # Define clauses for string array
    clauses.extend(string_constraints(strings, binary_strings))

    # closestString stores the currently closest string (binary encoded)
    closestString = lambda l, k: vpool.id('closestString{0}-{1}'.format(l, k))

    for l in range(L):
        cnf = CardEnc.equals(lits=[closestString(l, k) for k in range(4)], encoding=EncType.pairwise, bound=1)
        clauses.extend(cnf.clauses)

    # hammingDistances stores whether there is a distance between string n and the currently CS at position l
    # It might happen that too many entries are true, but since it is a satisfaction problem it does not matters
    hammingDistances = lambda n, l: vpool.id('hammingDistances{0}-{1}'.format(n, l))

    # Put distances into matrix
    for n in range(N):
        for l in range(L):
            for k in range(4):
                # When string and the current closest string are not the same at a position
                # then the hamming distance has to increase
                a = binary_strings(n, l, k)
                b = closestString(l, k)
                c = hammingDistances(n, l)
                # (c = 0) iff (a == b)
                # (a v b v c) /\ (-a v -b v c) /\ (a v -b v -c) /\ (-a v b v -c)
                clauses.extend([[-a, b, c], [a, -b, c]])

    # Constraint: All hamming distances smaller than d
    for n in range(N):
        cnf = atmost_all_bounds(lits=[hammingDistances(n, l) for l in range(L)], bound=d[n])
        clauses.extend(cnf)

    g = Glucose3()
    for clause in clauses:
        g.add_clause(clause)

    if g.solve():
        # print("This instance is solvable")
        model = g.get_model()

        cs = ""
        for l in range(L):
            if closestString(l, 0) in model:
                cs += "A"
            if closestString(l, 1) in model:
                cs += "C"
            if closestString(l, 2) in model:
                cs += "G"
            if closestString(l, 3) in model:
                cs += "T"
        # print(cs, "is the currently closestString")
        slimClosestString.value = cs
    else:
        slimClosestString.value = ""


# Input: length
# Output: Outputs a random String of given length, containing of the letters A, C, G, and T
def get_random_string(length):
    # choose from all lowercase letter
    letters = ["A", "C", "G", "T"]
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


# Input: Currently closest Sting, List of Strings, mode ("best" or "random")
# Output: Indices of substring which will be used for SLIM
def choose_subsequence(cs, strings, mode, length, L):
    prob = [1.0] * len(cs)

    if mode == "best":
        distances = evaluate(cs, strings, False)
        ind = distances.index(max(distances))
        for i in range(len(cs)):
            if not cs[i] == strings[ind][i]:
                prob[i] = 100.0

    elif mode == "random":  # random
        prob_sum = sum(prob)
        # Do nothing

    else:
        raise TypeError(f'Mode "{mode}" is invalid.')

    prob_sum = sum(prob)
    prob = [el / prob_sum for el in prob]

    indices = np.random.choice(range(len(cs)), length, replace=False, p=prob)
    return indices.tolist()



# Input: List of strings and list of indices which should be extracted
# Output: List of substrings depending on given indices, as well as strings without the extracted substrings
def extractSubsequences(cs, strings, indices):
    subsequences = []
    subCS = []
    for string in strings:
        subsequences.append([string[i] for i in indices])

    subCS = [cs[i] for i in indices]

    return subCS, subsequences


if __name__ == '__main__':
    start = time.time()

    d = 0  # Maximal distance
    l = 10   # default length of local instances
    iterations = -1  # number of SLIM iterations
    localTimeout = 60    # local timeout
    outfile = ""

    if args.d is not None:
        d = args.d

    if args.l is not None:
        l = args.l

    if args.iterations is not None:
        iterations = args.iterations

    if args.localTimeout is not None:
        localTimeout = args.localTimeout

    if args.strings is None:
        raise TypeError(f'You need to give the name of a file with all input strings')

    if args.closestString is None:
        raise TypeError(f'You need to give the name of a file with the currently closest string')

    if args.outfile is None:
        raise TypeError(f'You need to give the name of a output file for the solution')
    else:
        outfile = args.outfile

    log_name = str(Path(args.strings).stem) + "_l" + str(l) + ".log"

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(filename=log_name, encoding='utf-8', mode='w')
    root_logger.addHandler(handler)

    # Read input strings
    strings = []

    f = open(args.strings, "r")
    next(f)     # Skip first line
    for line in f:
        strings.append(line.strip())
    f.close()

    equal_lengths(strings)

    N = len(strings)
    L = len(strings[0])

    random.seed(10)

    # initial solution
    f = open(args.closestString, "r")
    current_cs = f.readline().strip()
    f.close()

    currentDistances = evaluate(current_cs, strings, False)

    curr_d = max(evaluate(current_cs, strings, False))
    logging.info("Original distance: {curr_d}".format(curr_d=curr_d))

    iter = 0
    while iter < iterations or iterations == -1:
        # print("--------------------------------")

        # Get start and end index of substring
        indices = choose_subsequence(current_cs, strings, "best", l, L)
        # print("Substring indices: ", indices)

        # Get substrings dependent on indices
        subCS, subsequences = extractSubsequences(current_cs, strings, indices)

        localDistances = evaluate(subCS, subsequences, False)
        remainingDistances = [a - b for a, b in zip(currentDistances, localDistances)]


        # Calculate the maximal distances for SLIM
        maxDist = max(remainingDistances)
        local_distances = [maxDist - remDist for remDist in remainingDistances]

        # SLIM
        solved = False
        while not solved:
            local_cs = Manager().Value(c_char_p, "")
            p = Process(target=slim, name="Slim", args=(subsequences, local_distances, local_cs))
            p.start()

            p.join(localTimeout)

            if p.is_alive():
                p.terminate()
                p.join()

            if local_cs.value == "":
                local_distances = [x + 1 for x in local_distances]

            else:
                new_cs = current_cs
                for i in indices:
                    new_cs = new_cs[:i] + local_cs.value[indices.index(i)] + new_cs[(i + 1):]
                solved = True

        localDistances = evaluate(local_cs.value, subsequences, False)
        currentDistances = [a + b for a, b in zip(remainingDistances, localDistances)]
        logging.info("current distances: {curr_dist}".format(curr_dist=currentDistances))
        temp = max(currentDistances)

        if temp <= curr_d:
            current_cs = new_cs

        if temp < curr_d:
            curr_d = temp
            end = time.time()
            logging.info("curr_d: {curr_d} - time: {time}".format(curr_d=curr_d, time=end - start))

            with open(outfile, 'w') as f:
                f.write(current_cs)
                f.write("\n\n")
                f.write("hamming distance: ")
                f.write(str(hamming_distance(strings, current_cs)))

        if curr_d <= d:
            break

        if iterations != -1:
            iter = iter + 1

    print("The closest string after SLIM with hamming distance of", str(hamming_distance(strings, current_cs)), "was written into the file", outfile)
